/*
 * H E A D E R FILE: Data structures templates.
 *
 *  written by: Async
 *  ver. 1.0
 *  HF of dstruct.cpp
 */

//#include <stdio.h>
//#include "gems.h"

/*
Contains:

  data struct(class)     |   status
  -----------------------+---------------
  FIFO                   |   N/A
  LIFO                   |   N/A
  Forward Linear List    |   0.5
  Bidirect Linear List   |   N/A
*/

#define MAX_FLL_CELLS 0xFFF

/* syntax:
  cX         ... class X
  sX         ... struct X
  aX         ... abstract class
  tX         ... template

*/

/**************************************************************************
 ***  Forward Linear List  ***
 **************************************************************************/

// Forward Linear List Cell
typedef struct sFllCell{
  sFllCell* next;                        // next cell
};

/*___________________________________________________________________________

                            Forward Linear List Base
____________________________________________________________________________*/


class cFllBase{
  private:
    sFllCell* base;                    // first cell
    sFllCell* current;                 // current cell pointer
    long int inx;                      // current cell index
    long int cs;                       // number of cells
    cGems* gems;
    int erase(sFllCell* c);
  public:
    int adds(sFllCell*);               // add cell to start (base)
    int adde(sFllCell*);               // add cell to end
    int insert(sFllCell*);             // insert at current position (liInx)
    sFllCell* pick(void);              // pick (get&remove) at current pos.
    sFllCell* picks(void);             // pick (get&remove) at start (base)
    sFllCell* picke(void);             // pick (get&remove) at the end

    sFllCell* browse(long int);        // list (read and go on N iSt)
    /* read current cell */
    sFllCell* read(void);
    /* read cell @ start (base) */
    sFllCell* reads(void);
    /* set current index (point to ...) */
    int  index(long int i);
    /* return current index */
    long int index(void);
    /* return size (number of stored cells) */
    long int cells(void);
    /* free memory of all list */
    int erase(void);
    /* constructor */
    cFllBase(cGems* g);
    /* destructor */
    ~cFllBase();
};

/****************************************************************************
 * FLL Template:
 ****************************************************************************/

template<class T> class tFll: private cFllBase{
  public:
    tFll(cGems* g):cFllBase(g){ };
    ~tFll(){ };
    int adds(T* c) { return cFllBase::adds(c); }
    T* reads(void) { return (T*)cFllBase::reads(); }
    T* browse(long int steps) { return (T*)cFllBase::browse(steps); }
    int index(long int i) { return cFllBase::index(i); }
    long int index(void) { return cFllBase::index(); }
    long int cells(void) { return cFllBase::cells(); }
};


/****************************************************************************
 * container template
 ****************************************************************************/

#define MAXINT 0xFFF
#define MAXPUFF 0xFF
#define MININT 0x8000

template<class T>
class tContainer {
  private:
    T* buf;        // buffer
    int size;      // allocated mem (max# of cells in container)
    int top;       // index of top of buffer
    int index;
    int puffstep;  // <0 enables "blowing" of container (reallocation)
    cGems* gems;   // global error messaging system
  public:
    // call this: puff enabled
    tContainer(int s, int ps, cGems* g);
    // call this: blowing disabled (default)
    tContainer(int s, cGems* g);
    ~tContainer();
    int max(void);
    int stored(void);
    int full(void);
    int empty(void);
    int over(void);
    int end(void);
    void zeroindex(void);
    int puffable(void);
    void add(T cell);
    /* write to position */
    void write(int inx, T cell);
    /* write from position */
    T read(int inx);
    /* remove top */
    T remove(void);
    T browse(void);
};



/*  T E M P L A T E
 *  I M P L E M E N T A T I O N:  Data Structures
 *
 *  written by: Async
 *  ver. 1.0
 *  IF of dstruct.cpp
 */



/*___________________________________________________________________________

                            Forward Linear List Base
____________________________________________________________________________*/


/* read current cell */
inline sFllCell* cFllBase::read(void){ return current; }

/* read cell @ start (base) */
inline sFllCell* cFllBase::reads(void){
  PUSH("sFllCell read from start");
  protect(base){ POP; return base; }
  POP; return NULL;
};

/* set current index (point to ...) */
inline int cFllBase::index(long int i){
  PUSH("cFllBase set index");
  if (cs<0) { POP; return 0; }
  rprotect(0,i,cs){
    sFllCell* carrige = base;
    inx = i;
    for (long int j = 0; j < i; j++) {
      protect(carrige){
        carrige = carrige->next;
      }
    }
    protect (carrige){ current = carrige; POP; return 0; };
    if (!carrige) { gmsgi("count:",cs); gmsgi("index:",i); }
  }
  POP;
  return -1;
}

/* return current index */
inline long int cFllBase::index(void) { return inx; }

/* return size (number of stored cells) */
inline long int cFllBase::cells(void) { return cs; }

/* free memory of all list */
inline int cFllBase::erase(void){
  PUSH("cFllBase erase");
  if (base){
    success(erase(base));
    base = NULL;
  }
  POP;
  return 0;
}

/* constructor */
inline cFllBase::cFllBase(cGems* g){
  gems = g;
  GOK(gems); //ten chcipak neumi tohle: GOK((gems=g));
  PUSH("cFllBase construct");
  inx = -1; cs = 0; base = current = NULL;
  POP;
  return;
};

/* destructor */
inline cFllBase::~cFllBase(void){
  PUSH("cFllBase destruct");
  success(erase());
  //gmsg("destroying");
  if (base) gmsg("memory not deallocated, erase() failed!");
  POP;
  return;
};

/* add cell to start (base) */
inline int cFllBase::adds(sFllCell* c){
  PUSH("cFllBase adds");
  //protect(c){
    c->next = base;
    base = c;
    rprotect(-1,inx,MAX_FLL_CELLS){
      rprotect(0,cs,MAX_FLL_CELLS+1){
        inx=0; cs++;
      }
    }
    POP;
    return 0;
  //}
  POP;
  return -1;
};

/* add cell to end */
inline int cFllBase::adde(sFllCell*){

};

/* insert at current position (liInx) */
inline int cFllBase::insert(sFllCell*){

};

/* pick (get&remove) at current pos. */
inline sFllCell* cFllBase::pick(){

};

/* pick (get&remove) at start (base) */
inline sFllCell* cFllBase::picks(){

};

/* pick (get&remove) at the end */
inline sFllCell* cFllBase::picke(){

};

/* list (read and go on N iSt) */
inline sFllCell* cFllBase::browse(long int steps){
  PUSH("cFllBase browse");
  sFllCell* ret = current;
  if (inx==-1) { POP; return NULL; }
  if (current) {
    for (long int j = 0; j < steps; j++){
     current = current->next;
     if (!current) break;
     if (current) inx++;
    }
  }
  POP;
  return ret;
};


/* free memory of all list */
inline int cFllBase::erase(sFllCell* c){
  // don't use PUSH/POP due to recursive call
  protect(c){
    if (c->next != NULL) erase(c->next);
    delete c;
    return 0;
  }
  return -1;
};


/****************************************************************************
 * container template
 ****************************************************************************/

/* constructor - container blowing enabled */
template<class T> inline tContainer<T>::tContainer(int s, int ps, cGems* g){
  gems = g;
  GOK(gems);
  PUSH("container construct1");
  top = -1;
  index = 0;
  puffstep = 0;
  rprotect(0,ps,MAXPUFF){ puffstep = ps; }
  buf = NULL;
  rprotect(0,s,MAXINT){ buf = new T[size = s]; }
  POP;
  return;
}

/* constructor - blowing disabled (default) */
template<class T> inline tContainer<T>::tContainer(int s, cGems* g){
  gems = g;
  GOK(gems);
  PUSH("container construct2");
  top = -1;
  puffstep = 0;
  index = 0;
  buf = NULL;
  rprotect(0,s,MAXINT){ buf = new T[size = s]; }
  POP;
  return;
}

/* destructor */
template<class T> inline tContainer<T>::~tContainer(){
  PUSH("container destruct");
  protect(buf){ delete[] buf; }
  POP;
}

template<class T> inline int tContainer<T>::max(void){ return size; }
template<class T> inline int tContainer<T>::stored(void){ return (top+1); }
template<class T> inline int tContainer<T>::full(void){ return ((top+1) >= size); }
template<class T> inline int tContainer<T>::empty(void){ return (top < 0); }
template<class T> inline int tContainer<T>::over(void){ return (index > top); }
template<class T> inline int tContainer<T>::end(void){ return (index == top); }
template<class T> inline void tContainer<T>::zeroindex(void){
  index = 0;
  //gmsgi("ZERO Index:",index);  return;
}

template<class T> inline int tContainer<T>::puffable(void) { return puffstep; }

template<class T> inline T tContainer<T>::remove(void){
  rprotect(0,top,size-1){
    return buf[top--];
  }
  return NULL;
}

template<class T> inline T tContainer<T>::browse(void){
  index++;
  return read(index-1);
}


/* add to top of container */
template<class T> void tContainer<T>::add(T cell){
  PUSH("container add");
  if ((top<(size-1))||(!puffstep)){
    // blowing disabled
    rprotect(-1,top,size){ protect(buf){ buf[++top] = cell; }; }
  } else {
    // blowing enabled
    gmsg("PUFFING");
    rprotect(-1,size+puffstep,MAXINT){
      protect(buf){
        size += puffstep;
        T* tmp = new T[size];
        // copy data
        protect(tmp){
          for (int i = 0; i <= top; i++) tmp[i] = buf[i];
        }
        // free mem
        delete[] buf;
        // new bigger buffer
        buf = tmp;
        buf[++top] = cell;
      }
    }
  }
  POP;
  return;
};


/* write to position */
template<class T> void tContainer<T>::write(int inx, T cell){
  PUSH("container write");
  rprotect(0,inx,size-1){
    if (inx <= top) {
      protect(buf){  buf[inx] = cell; }
    } else gmsgi("Attempt at writing to free space:",inx);
  }
  POP;
  return;
}

/* read from position */
template<class T> T tContainer<T>::read(int inx){
  PUSH("container read");
  rprotect(0,inx,size-1){
    if (inx <= top) {
      protect(buf){ POP; return buf[inx];}
    } else gmsgi("Attempt at reading from free space:",inx);
  }
  T blank;
  POP;
  return blank;
}


  #undef MAXINT
  #undef MAXPUFF
  #undef MININT

