/*
 * H E A D E R FILE: Global Error Messaging System
 *
 *  ver. 1.0
 */

#undef NULL
#define NULL 0

#define DEBUGLEVEL 4

#define MX_MSG_LEN  250                /* max. message lenght */
#define MX_CSEG_LEN  50                /* max.  code segment name */

#define GemsPtr     0
#define GemsInt     1
#define GemsFloat   2
#define GemsChar    3
#define GemsString  4

#ifndef DEBUGLEVEL
  #define DEBUGLEVEL 0
#endif

#if (DEBUGLEVEL<0 || DEBUGLEVEL>5)
  #error DEBUGLEVEL out of range!
#endif

#if (DEBUGLEVEL==1)
  #define DEBUG
  #define DEBUG_USER_MSG
  #define DEBUG_PUSHPOP
  #define DEBUG_GOK
#endif

#if (DEBUGLEVEL==2)
  #define DEBUG_USER_MSG
  #define DEBUG_POINTER
  #define DEBUG_RANGEPROTECT
  #define DEBUG_PUSHPOP
  #define DEBUG_GOK

#endif

#if (DEBUGLEVEL==3)
  #define DEBUG_USER_MSG
  #define DEBUG_POINTER
  #define DEBUG_FRES_CHECK
  #define DEBUG_RANGEPROTECT
  #define DEBUG_RANGE
  #define DEBUG_PUSHPOP
  #define DEBUG_GOK

#endif

#if (DEBUGLEVEL==4)
  #define DEBUG_USER_MSG
  #define DEBUG_POINTER
  #define DEBUG_STRINGS
  #define DEBUG_RANGEPROTECT
  #define DEBUG_RANGE
  #define DEBUG_FRES_CHECK
  #define DEBUG_PUSHPOP
  #define DEBUG_GOK
#endif

#ifdef DEBUG_PUSHPOP
  #define PUSH(name) gems->sgpush(name);
  #define POP gems->sgpop();
#else
  #define PUSH(name)
  #define POP
#endif

#ifdef DEBUG_GOK
  #define GOK(g) if ((g) == NULL) { \
    printf("FATAL ERROR GEMS POINTER NULL!",__FILE__,__LINE__);\
    exit(-1);}
#else
  #define GOK(g)
#endif
/* DEBUG - user messages only */
#ifdef DEBUG_USER_MSG
 // use macro gmsg NOT msg to broadcast message to gems!
 #define gmsg(m) gems->msg(__FILE__,__LINE__,m)
 #define gmsgi(m,i) gems->msg(__FILE__,__LINE__,m,i)
 #define gmsgs(m,s) gems->msg(__FILE__,__LINE__,m,s)
#else
 #define gmsg(m)
 #define gmsgi(m,i)
 #define gmsgs(m,s)
#endif

/* DEBUG - pointer protection */
#ifdef DEBUG_POINTER
 // protection against read/write from/to null pointer
 // usage: protect(p){ ... protected  code ... }
 #define protect(p) if((p) == NULL) gems->msg(__FILE__,__LINE__,"Pointer\
 caught as NULL!, Access denied.");  else
#else
 #define protect(p)
#endif

/* DEBUG - function results checking */
#ifdef DEBUG_FRES_CHECK
 // when function returns nonzero value = error value.
 #define success(f) { int r=f; if (r) gems->msg(__FILE__,__LINE__,"\
 Function returned nonzero value",r); }
#else
 #define success(f)
#endif

/* DEBUG - range checking */
#ifdef DEBUG_RANGE
 #define range(lower,x,upper) if ((x)<(lower) || (x)>(upper)) gems->msg(\
  __FILE__,__LINE__,"Range check error:",x)
#else
 #define range(lower,x,upper)
#endif

/* DEBUG - range checking, protection */
// usage: rprotect(lower,x,higher,gems pointer){ ... protected  code ... }
#ifdef DEBUG_RANGEPROTECT
 #define rprotect(lower,x,upper) if ((x)<(lower) || (x)>(upper)) gems->msg(\
 __FILE__,__LINE__,"Range check error. Protected process skipped.:",x); else
#else
 #define rprotect(lower,x,upper)
#endif

/* DEBUG - strings length checking */
// usage: sufficient(size_d,s,gems pointer){ ... protected  code ... }
#ifdef DEBUG_STRINGS
 #define sufficient(size_d,s) if ((size_d)<(strlen(s)+1)) gems->msg(__FILE__\
  ,__LINE__,"Destinantion array or string: Insufficient space. \
  Protected process skipped."); else
#else
 #define sufficient(size_d,s)
#endif



class cGems {
  private:
    int iMBsize;                       /* buffer size */
    int iWsize;                        /* watch stack size */
    int iCSsize;                       /* "code segment" name stack */
    int wp;                            /* watch pointer */
    int csp;                           /* cseg pointer */

    typedef char mstr[MX_MSG_LEN+1];   /* max. message lenght */
    typedef char cstr[MX_CSEG_LEN+1];  /* max. code segment name */
    mstr* mbuf;                        /* message buffer - queue */
    cstr* cseg;                        /* cseg stack */
    void** watch;                       /* watch stack */

    int trace;                         /* stop whenever comes new message
                                          and "screen focus" */
  public:
    void echo(void);
    void traceOn(void){ trace = 1; }   /* switch ON trace */
    void traceOff(void){ trace = 0; }  /* switch OFF trace */
    int brk(int iln);                  /* call break @ line #iln */

    int listing(void);                /* write all buffer on screen */

    void msg(char* fn, int ln, char* m);/* add new message to queue */
    /* msg with integer parameter */
    void msg(char* fn, int ln, char* m, int x);
    /* msg with string parameter */
    void msg(char* fn, int ln, char* m, char* x);

    int sgpush(char* segName);         /* add new code segment to "cseg"
                                          stack - call in every procedure
                                          when proc. starts it's useful to
                                          locate errors */

    int sgpop(void);                     /* pop cseg stack - call while
                                          finalizing procedure (before return)
                                                                            */

                                       /* constructor */
    cGems(int icBsize, int icWsize, int icCSsize);
                                       /* destructor */
    ~cGems(void){
      delete[] mbuf;
      delete[] watch;
      delete[] cseg;
    }

};

// END OF HEADER FILE
