
                        /* General Language Database */

// written by: Async, Peter. http://asyncbrain.baf.cz

/*
 * basic clases:
 *           - cRadix
 *           - cCompound as set of words
 *           - cClause as set of words
 *           - relation between radix/composite/clause
 */

/* nomenclature:
 *  sName    - typedef_ined structure
 *  eName    - enumeration
 *  aName    - abstract class
 *  cName    - class
 *  tName    - template
 *
 */
#include   <stdio.h>
#include   <ctype.h>
#include   <strings.h>
#include   <string.h>
#include   "gldtype.cc"


/* relation */
class cRelation{
  public:
    RELATIONID rid;      // relation id number
    sReference rex;      // relation extension (cClause)
    sReference ref;      // relation ``conncetion''
};

/* aJay provide bidirect query    of   KeyList<->GLDCell
 *                                     Lexicon<->GLDCell
 *                                     KeyList<->Lexicon
 */
class aJay{
  protected:
    aJay* backquery;
  public:
   aJay(aJay* ini) { backquery = ini; }
   ~aJay(void){};
   /* text search */
   virtual aGLDCell* texsea(char* s, TSM scope) = 0;
   /* reference search, backqry enables call "back query" if ref key!=ref */
   virtual aGLDCell* refsea(sReference ref) =  0;
   virtual HSKEY reftokey(sReference ref) = 0;   
   //don't use!virtual void seares(void) = 0;
   aJay* backqry(void){ return backquery; }
};

/* Class aGLDCell is the general database abstract class because of
 * compatibility radix, compound, clause.
 */

class aGLDCell : public sFllCell{
  protected:
    cRelContainer* relations;
    sReference cellref;                       /* cell ID,lang, type */
    cGems* gems;
    aJay* backqry;
  public:
    /* constructor (direct fill with relations) */
    aGLDCell(sReference r, cRelContainer* fill, aJay* bq, cGems* g){
      GOK(g); gems = g;
      PUSH("aGLDCell construct1");
      relations = NULL;
      cellref = r;
      backqry = bq;
      protect(fill){
        // create copy of filling container of relations
        relations = new cRelContainer(fill->stored(),g);
        // cpy
        fill->zeroindex();
        while (!relations->full()) relations->add(fill->browse());
      }
      POP; return;
    };
    /* constructor (default size, unknown content)
     * tContainer can automatically puff up to necessary size
     */
    aGLDCell(sReference r,  aJay* bq, cGems *g){
      GOK(g); gems = g;
      PUSH("aGLDCell construct2");
      cellref = r; //.hk = r.hk; cellref.id = r.id; cellref.lang = r.lang;
      backqry = bq;
      relations = new cRelContainer(GLDCELL_REL_DEFSIZE,GLDCELL_REL_PUFF,g);
      // NULL pointer => backqry not installed yet.
      POP; return;
    }
    /* destructor */
    ~aGLDCell(){
      // destroy container of relations
      PUSH("aGLDCell destruct");
      protect(relations) { delete relations; }
      POP; return;
    };
    /* returns reference of cell */
    sReference ref(void){ return cellref; }
    /* modify reference of cell */
    void ref(sReference r){
      cellref.id = r.id;
      cellref.lang = r.lang;
      cellref.hk = r.hk;
      return;
    }
    /* set back query if NULL */
    int setbackqry(aJay* bq){
      if (!backqry) { backqry = bq; return 0; }
      return -1;
    }

    /* returns formated radix/composite/clause */
    virtual void text(char* str, TEXTFORMAT tf, int chars) = 0;
    virtual char* text(TEXTFORMAT tf) = 0;
    virtual void print(void) = 0;
    /* self-identification */
    virtual GLDCELLTYPE gldctype(void) = 0;
    /* add relation */
    int addrel(cRelation rel){
      if (!relations->puffable()) return -1; // add disabled
      //gmsg("rel added");
      relations->add(rel); return 0;     
    }
    cRelRefBag* getrelation(void){
      PUSH("aGLDCell getrelations");
      if (relations->stored()<=0) { POP; return NULL; }
      cRelRefBag* bag = new cRelRefBag(relations->stored(),gems);
      // copy
      relations->zeroindex();
      while (!bag->full()) bag->add(relations->browse());
      POP; return bag;
    }
    virtual aGLDCell* clone(void) = 0;
};

/* GLDCell key list
 * keyed list contains GLDCells only with spec. key (1st character)
 */

class cKeyList:public aJay{
  private:
    HSKEY key;
    GLDCELLID counter;// id counter
    cChainFll* chain;
    cGems* gems;
  public:
    /* constructor */
    cKeyList(HSKEY k,aJay* l,cGems* g):aJay(l){
      gems = g; GOK(g);
      PUSH("cKeyList construct");
      //~ connect list to lexicon - back query call
      //~backqry = NULL;
      //~protect(l){ backqry = l; };
      // set hash key:
      key = k;
      counter = NULL_ID;
      chain = new cChainFll(g);
      POP; return;
    };
    /* destructor */
    ~cKeyList(){
      PUSH("cKeyList destruct");
      //gmsgi("deleting chain HK:",key);
      protect(chain) { POP; delete chain; return; };
      POP;
      return;
    };
    /* add new cell */
    int add(aGLDCell* c){
      PUSH("cKeyList add");
      protect(chain){
        protect(c){
          // put back query call:
          //success(c->setbackqry(this));
          c->setbackqry(this);
          // put ID:
          sReference r = c->ref();
          if (r.id== NULL_ID) r.id = ++counter; else {
            if (r.id > counter) counter = r.id;
            if (r.id<= counter) {
              gmsgi("ID of cell <= counter, database error! id:",r.id);
              POP; return -1;
            }
          }
          //~ gmsgi("ID:",r.id);
          c->ref(r);
          //~ gmsgi("COUNTER:",counter);
          chain->adds(c);
        }
      POP; return 0;
      }
      //aGLDCell* tmp = chain->reads();
      //protect(c){ ~ gmsgi(tmp->text(TF_PLAIN,NULL),key); }
      POP; return 0;
    }
    /* text search, backqry enables call "back query" */
    aGLDCell* texsea(char* s, TSM scope);
    /* reference search, backqry enables call "back query" if ref key!=ref */
    aGLDCell* refsea(sReference ref);
    /* reset current index (goto start of chain) */
    void seares(void) { chain->index(0); }
    HSKEY reftokey(sReference ref){ return backquery->reftokey(ref); }
    long int index(void){ return chain->index(); }
    void printall();
};

void cKeyList::printall(){
  if (chain->cells()>0){
    chain->index(0);
    //printf("\n");    
    while (aGLDCell* cell=chain->browse(1)) { 
      if (cell) {
        char* s = cell->text(TF_PLAIN);
        printf("[%s]",s); delete[] s;
      }	 else {
        printf("[NULL]");
      }
    }  
  }
};



/* lexicon - contains all radixes, compounds, clauses of one language */
class cLexicon: public aJay{
  private:
    cKeyListCont* lexicon;
    HSKEY hash(char* s);     // generate hash from string
    HSKEY hasho(short int o); // order of "hashes"
    sLexLabel lexlabel;
    cGems* gems;
  public:
    /* constructor */
    cLexicon(sLexLabel ilexlabel,aJay* bq, cGems* g):aJay(bq){
      GOK(g); gems = g;
      PUSH("cLexicon construct");
      // label lexicon
      lexlabel.name = NULL; lexlabel.kwrd = NULL;
      rprotect(0,ilexlabel.id,MAX_LEXICONS){
        int namelen = strlen(ilexlabel.name);
        rprotect(0,namelen,LEX_NAME_LEN){
          int kwrdlen = strlen(ilexlabel.kwrd);
          rprotect(0,kwrdlen,LEX_KEYW_LEN){
            lexlabel.name = new char[namelen+1];
            lexlabel.kwrd = new char[kwrdlen+1];
            lexlabel.id = ilexlabel.id;
            strcpy(lexlabel.name,ilexlabel.name);
            strcpy(lexlabel.kwrd,ilexlabel.kwrd);
          }
        }
      }
      // create pointer's table
      lexicon = new cKeyListCont(HS_KEYS,gems);
      // allocate mem for cell chains (key lists)
      register short int i = 0;
      while (i<HS_KEYS){
        lexicon->add(new cKeyList(hasho(i),NULL,gems));
        i++;
      }
      POP; return;
    }
    ~cLexicon(){
      PUSH("cLexicon destruct");
      cKeyList* tmp;
      while (!lexicon->empty()) {
        protect(tmp = lexicon->remove()){ delete tmp; }
      }
      //gmsg("Freeing lexicon:");
      delete lexicon;
      protect(lexlabel.name){ delete[] lexlabel.name; }
      protect(lexlabel.kwrd){ delete[] lexlabel.kwrd; }
      //~ gmsg("done.");
      POP; return;
    }
    /* search reference */
    aGLDCell* refsea(sReference ref);
    /* search text */
    /* text search, backqry enables call "back query" */
    aGLDCell* texsea(char* s, TSM scope);
    //...
    /* book, inscribe GLD cell */
    int inscribe(aGLDCell* cell);
    HSKEY reftokey(sReference ref){ return ref.hk; }
    void seares(){};
    /* lexicon label: laguage name, laguage keyword, #id */
    sLexLabel label(void){ return lexlabel; }
    char* name(void){ return lexlabel.name; }
    char* kwrd(void){ return lexlabel.kwrd; }
    LANGUAGEID id(void) { return lexlabel.id; }
    void printall();
};

void cLexicon::printall(){
  for (int hk=0; hk<HS_KEYS; hk++){
    cKeyList* kl = lexicon->read(hk);
    printf("HK:%d",hk);
    kl->printall();
    //printf("\n");
  }    
}
/* generate hash from string */
HSKEY cLexicon::hash(char* s){
  PUSH("cLexicon hash1");
  HSKEY hk = -1;
  protect(s){
    int len = strlen(s);
    if (len){
      char ch1 = tolower(s[0]);
      if ((ch1>='a')&&(ch1<='z')){
        if (len>1) {
          char ch2 = tolower(s[1]);
          if (ch2>'l') hk = (HSKEY)(ch1-'a'+(HS_KEYS-1)/2);
        }
        POP; return (HSKEY)(ch1-'a');
      } else hk = HS_KEYS-1;
    }
  }
  rprotect(0,hk,HS_KEYS-1){ POP; return hk; return 0; }
  POP;
  return -1;
};

/* order of "hashes" */
HSKEY cLexicon::hasho(short int o){
  PUSH("cLexicon hash2");
  rprotect(0,o,HS_KEYS-1){ POP; return o; };
  POP; return 0;
};

int cLexicon::inscribe(aGLDCell* cell){
  PUSH("cLexicon inscribe");
  protect(cell){
    sReference r = cell->ref();
    cKeyList* chain;
    // initialize HK if not known! (compiler support)
    rprotect(-1,r.hk,HS_KEYS){
      if (r.hk==NULL_HK) {
        char str[3];
        cell->text(str,TF_PLAIN,2);
        //gmsg(str);
        HSKEY hk = hash(str);
        chain = lexicon->read(hk);
        // include to ref.id hash key!
        // Lexicon gives hk
        r.hk = hk;
        cell->ref(r);
      } else { chain = lexicon->read(r.hk); } // cell already has HKey
    }
    // add cell
    //char* s = cell->text(TF_PLAIN); gmsg(s); delete[] s;
    protect(chain){ chain->add(cell); POP; return 0; }
  }
  POP; return -1;
};

/* search reference */
aGLDCell* cLexicon::refsea(sReference ref){
  PUSH("cLexicon refsea");
  HSKEY hk = reftokey(ref);
  rprotect(0,hk,HS_KEYS-1){    
     cKeyList* chain = lexicon->read(hk);
     protect(chain){
      //~ gmsgi("INDEX:",chain->index());
       aGLDCell* cell = chain->refsea(ref);      
       POP; return cell;
     }
  }
  POP; return NULL;
};

/* text search, backqry enables call "back query" */
aGLDCell* cLexicon::texsea(char* s, TSM scope){
  PUSH("cLexicon texsea");
  HSKEY hk = hash(s);
  rprotect(0,hk,HS_KEYS-1){
    cKeyList* chain = lexicon->read(hk);
    protect(chain){
      // ## this is not required, but used to achieve compatiblity with "aJay":
      // satisfing:
      aGLDCell* cell = chain->texsea(s,scope);
      //if (cell) gmsg(cell->text(TF_PLAIN));
      POP; return cell;      	       
    }
  };
  POP; return NULL;
};



/*cRadix - base of all words */
class cRadix:public aGLDCell{
  private:
    char* radix;               /* radix e.g. cat, |ness, pre|, |ize */
  public:
    /* known relations */
    cRadix(sReference iref, char* ir, cRelContainer* fill, aJay* bq, cGems *g):
      aGLDCell(iref,fill,bq,g){
      GOK(g); //gems = g;
      PUSH("cRadix construct1");
      radix = NULL;
      //gmsg("radix alloc1");
      long int len = strlen(ir);
      rprotect(0,len,MAX_RADIX_LEN){
        radix = new char[len+1];
        strcpy(radix,ir);
      }
      POP; return;
    };
    /* unknown relations */        
    cRadix(sReference ref, char* ir, aJay* bq,cGems *g):aGLDCell(ref,bq,g){      
      GOK(g); //gems = g;
      PUSH("cRadix construct2");
      radix = NULL; 
      long int len = strlen(ir);      
      rprotect(0,len,MAX_RADIX_LEN){
        radix = new char[len+1];
        strcpy(radix,ir);
      }
      POP; return;
    };
    ~cRadix(){
      protect(radix){ delete[] radix; };
    };
    /* returns reference of cell */
    //sReference reference(void){ return cellref; }
    /* formated string output with limited strlen, str must be allocated
     * before call. This achieve high speed of search */
    void text(char* str, TEXTFORMAT tf, int chars){
      PUSH("cRadix text1");
      rprotect(1,chars,MAX_STRING_LEN){
        protect(str){
          str[0]='\0';	  
	  char* s = str;
	  char* r = radix;
	  int i = 0;
	  while ((*s++ = *r++) && ((++i)<chars)); *s = '\0';
          //strncpy(str, radix, chars+1); // nefunguje! nejaky posrany.	  
	  //gmsgi(str, chars);
        }
      }
      POP; return;
    };
    /* formated string output with whole length, str is automatically
     * allocated, YOU MUST FREE MEM! */
    char* text(TEXTFORMAT tf){
      PUSH("cRadix text2");
      char* str = new char[strlen(radix)+1];
      strcpy(str,radix);
      POP; return str;
    };
    void print(){ };
    /* self-identification */
    GLDCELLTYPE gldctype(void){ return GLD_CT_RADIX; }
    /* replicate current object */
    aGLDCell* clone(){ 
      //printf("clone Radix");    
      return new cRadix(cellref,radix,relations,backqry,gems);
     }
};

class cCompound:public aGLDCell{
  private:
    cRefContainer* comps;      /* compound set e.g. ill|ness, 
                                          computer|ize */
  public:
    /* known relations */
    cCompound(sReference ref, cRefContainer* c,
     cRelContainer* fill, aJay* bq,cGems *g):aGLDCell(ref,fill,bq,g){
      GOK(g); //gems = g;
      PUSH("cCompound construct1");
      //gmsg("comp alloc1");
      comps = NULL;
      rprotect(0,c->stored(),MAX_COMPOUND_CELLS){
        //gmsg("comp alloc2");      
        comps = new cRefContainer(c->stored(),g);
        c->zeroindex();
	//gmsg("filling");
        while (!comps->full()) comps->add(c->browse());
      }
      POP; return;
    };
    /* unknown relations */
    cCompound(sReference ref, cRefContainer* c, aJay* bq, cGems *g)
     :aGLDCell(ref,bq,g){
      GOK(g); //gems = g;
      PUSH("cCompound construct2");
      comps = NULL;
      rprotect(0,c->stored(),MAX_COMPOUND_CELLS){
        comps = new cRefContainer(c->stored(),g);
        c->zeroindex();
        while (!comps->full()) comps->add(c->browse());
      }
      POP; return;
    };
    ~cCompound(){
      PUSH("cCompound destruct");
      protect(comps){ delete comps; }
      POP; return;
    };
    /* returns reference of cell */
    //sReference reference(void){ return cellref; }
    /* formated string output with limited strlen, str must be allocated
     * before call. This achieve high speed of search */
    void text(char* str, TEXTFORMAT tf, int chars){
      PUSH("cCompound text1");
      if (!backqry) gmsg("Compound require back query call!"); else {
       rprotect(1,chars,MAX_STRING_LEN){
        protect(str){
          str[0]='\0'; //blank sring
          // assemble compoud from radixes:
          comps->zeroindex();
          // if text format bla|bla|ble required
          // allocate +(comps->stored()-1) chars.
          int rest = chars; // not required +comps->stored..
          char* radix = new char[chars+1];
          while ((!comps->over()) && (rest>0)){
            // get one radix:
            sReference cref = comps->browse();
            // search for radix:
            protect(backqry){
              aGLDCell* cell = backqry->refsea(cref);
              //~ gmsgi("cref.id",cref.id);
              //~ gmsgi("cref.hk",cref.hk);
              //~ gmsgi("cref.lang",cref.lang);
              // get content of radix + add:
              protect(cell){
                // check if cell is radix!
                if (cell->gldctype()!= GLD_CT_RADIX)
                  gmsg("Radix expected, compound can't contain compound "
                        "or clause, error in database!");
                else {
                  cell->text(radix,TF_PLAIN,rest);
                  //~gmsg(radix);
                  //char* s = str;
                  //char* r = radix;
                  //while ((*s++ = *r++)&&((radix-r)<=rest));
                  strncat(str,radix,rest);
                  rest -= strlen(radix);
                  // if text format bla|bla|ble required
                  // add '|' strcpy(str,"|");
                  // dont' rest++ !
                }
              }
            }
          }
          delete[] radix;
        }
       }
      }
      POP; return;
    };
    /* formated string output with whole length, str is automatically
     * allocated, YOU MUST FREE MEM! */ #
    char* text(TEXTFORMAT tf){
      PUSH("cCompound text2");
      //~gmsg("cCompound asked for text1");
      if (!backqry) gmsg("Compound require back query call!"); else {
        // allocate string container.
        int ncells = comps->stored();
        cStrContainer* scont = new cStrContainer(ncells,gems);
        // assemble compoud from radixes:
        comps->zeroindex();
        // if text format bla|bla|ble required
        int len = 0; // total lenght of contacted string
        while (!comps->over()){
          // get one radix:
          sReference cref = comps->browse();
          //gmsgi("REF HK:",cref.hk);
          //gmsgi("REF LANG:",cref.lang);
          //gmsgi("REF ID:",cref.id);
          // search for radix:
          aGLDCell* cell = backqry->refsea(cref);
          protect(cell){
            // check if cell is radix!
            if (cell->gldctype()!= GLD_CT_RADIX)
              gmsg("Radix expected, compound can't contain compound "
                    "or clause, error in database!");
            else {
              char* s = cell->text(TF_PLAIN);
              //~gmsg(s);
              len += strlen(s);
              scont->add(s);
            }
          }
        }
        char* outs = new char[len+1];
        // contact partial strings (radixes):
        scont->zeroindex();
        outs[0] = '\0'; // blank string
        for (int i = 0; i<ncells; i++) {
          char* tmp =  scont->browse();
          strcat(outs,tmp);
	  delete[] tmp;
	};
        delete scont;
        POP; return outs;
      }
      POP;
     // char* str = new char[strlen(radix)+1];
     // strcpy(str,radix);
      return NULL;
    };


    void print(){ };
    /* self-identification */
    GLDCELLTYPE gldctype(void){ return GLD_CT_COMPOUND; }
    aGLDCell* clone(){ 
      //printf("clone Compound");
      return new cCompound(cellref, comps, relations, backqry, gems);
    }
};

class cClause:public aGLDCell{
  private:
    cRefContainer* clause;      /* clause e.g. I am ever in doubt...
                                           */
  public:
    /* known relations */
    cClause(sReference ref, cRefContainer* cl,
     cRelContainer* fill, aJay* bq, cGems *g):aGLDCell(ref,fill,bq,g){
      GOK(g); //gems = g;
      PUSH("cClause construct1");
      clause = NULL;
      //gmsg("clause alloc1");
      rprotect(0,cl->stored(),MAX_CLAUSE_CELLS){
        clause = new cRefContainer(cl->stored(),g);        
	cl->zeroindex();
        while (!clause->full()) clause->add(cl->browse());
      }
      POP; return;
    };
    /* unkown relations */
    cClause(sReference ref, cRefContainer* cl, aJay* bq, cGems *g)
     :aGLDCell(ref,bq,g){
      GOK(g); //gems = g;
      PUSH("cClause construct2");
      //gmsg("clause construction...");
      clause = NULL;
      rprotect(0,cl->stored(),MAX_CLAUSE_CELLS){      
        clause = new cRefContainer(cl->stored(),g);        
	cl->zeroindex();
        while (!clause->full()) clause->add(cl->browse());
      }
      POP; return;
    };
    ~cClause(){
      PUSH("cClause destruct");
      protect(clause){ delete clause; }
      POP; return;
    };
    /* returns reference of cell */
    //sReference reference(void){ return cellref; }
    /* formated string output with limited strlen, str must be allocated
     * before call. This achieve high speed of search */
    void text(char* str, TEXTFORMAT tf, int chars);
    
    /* formated string output with whole length, str is automatically
     * allocated, YOU MUST FREE MEM! */
    char* text(TEXTFORMAT tf);    

    void print(){ };
    /* self-identification */
    GLDCELLTYPE gldctype(void){ return GLD_CT_CLAUSE; }
    aGLDCell* clone(){ 
      //printf("clone Clause");    
      return new cClause(cellref, clause, relations, backqry, gems);
    }
};

/* Librarian serves all lexicons */
class cLibrarian : aJay{
  private:
    cLexContainer* lib;
    cRelList* relist;
    cGems* gems;

    // unused, inherited from virtual Jay:
    aGLDCell* texsea(char* s, TSM scope){ return NULL; }
    aGLDCell* refsea(sReference ref){ return NULL; }
    HSKEY reftokey(sReference ref){ return -1; }

    /* relation keyword to relation ID */
    RELATIONID relkeytoid(char* kwrd);
    int relisunary(RELATIONID id){
      PUSH("cLibrarian relisunary");
      sRelation r = relist->read(id);
      POP; return (r.unary != 0);
    }

  public:
    cLexicon* chooselex(char* kwrd);
    cLexicon* chooselexref(sReference r);
    /* search - cLibrarins's */
    aGLDCell* searchone(sExpression e);
    aGLDCell* searchoneref(sReference r);
    /* construct */
    cLibrarian(int maxlex, int maxrel, cGems* g):aJay(NULL){
      GOK(g); gems = g;
      PUSH("cLibrarian construct");
      // allocate lexicon library:
      rprotect(0,maxlex,MAX_LEXICONS){
        lib = new cLexContainer(maxlex,gems);
        relist = new cRelList(maxrel,gems);
        // lexicons will add user.
      }
      POP; return;
    }
    /* destruct */
    ~cLibrarian(){
      PUSH("cLibrarian destruct");
      while (!lib->empty()) {
        cLexicon* tmp;
        tmp = lib->remove();
        protect(tmp) { /* gmsgs("freeing:", tmp->name());*/ delete tmp; };
      }
      POP; return;
    }
    /* add expression in language e.kwrd */
    aGLDCell* addradix(sExpression e);
    aGLDCell* addcompound(sExpression e);
    aGLDCell* addclause(sExpression e);
    /* add relation betwen 2 expressions/ cells */
    int addrelation(sExpression e1, sExpRelation r, sExpression e2);
    /* add lexicon - new language */
    int addlexicon(sLanguage lang);
    /* add new type of relation */
    int addrelation(sRelation rel);
    /* return ``bag of relation owned by cell'' */
    cRelationBag* getrelation(aGLDCell* cell);
    /* search for expression/cell, returns all with leading equ str. */
    cCellBag* search(sExpression e);
    void printall();
    int lexexist(char* kwrd){ if (chooselex(kwrd)!=NULL) return 1; else return 0; }
    int relexist(char* kwrd){ if (relkeytoid(kwrd)>=0) return 1; else return 0;  }
};

/* privates cLexicon */
inline cLexicon* cLibrarian::chooselex(char* kwrd){
  PUSH("cLibrarian chooselex");
  lib->zeroindex();
  while (!lib->over()) {
    cLexicon* lex = lib->browse();
    protect (lex){
      if (!strcmp(lex->kwrd(),kwrd)) { POP; return lex; }
    }
  }
  POP; return NULL;
};

inline cLexicon* cLibrarian::chooselexref(sReference r){
  PUSH("cLibrarian chooselexref");
  lib->zeroindex();
  while (!lib->over()) {
    cLexicon* lex = lib->browse();
    protect (lex){
      //gmsgi("LEXICON:",lex->id());
      if (lex->id()==r.lang) { POP; return lex; }
    }
  }
  POP; return NULL;
};

inline aGLDCell* cLibrarian::searchone(sExpression e){
  PUSH("cLibrarian searchone");
  cLexicon* lex = chooselex(e.kwrd);
  if (lex){
    aGLDCell* cell = lex->texsea(e.str,TSM_FS_EQU);
    POP; return cell;
  } else {
    gmsg("Invalid language, lexicon not found!");
    POP; return NULL;
  }
};

inline aGLDCell* cLibrarian::searchoneref(sReference r){
  PUSH("cLibrarian searchoneref");
  cLexicon* lex = chooselexref(r);
  if ((r.id==-1) || (r.hk==-1)) { POP; return NULL; }
  if (lex){
    aGLDCell* cell = lex->refsea(r);
    POP; return cell;
  } else {
    gmsgi("Invalid language, lexicon not found! languageID:",r.lang);
    POP; return NULL;
  }
};

inline RELATIONID cLibrarian::relkeytoid(char* kwrd){
  PUSH("cLibrarian relkeytoid");
  relist->zeroindex();
  if (relist->empty()){ POP; return -1; }
  RELATIONID id = 0;
  while (!relist->over()) {
    sRelation r = relist->browse();
    if (!strcmp(r.kwrd,kwrd)) { POP; return id; } // return ID
    id++;
  }
  POP; return -1; //not found
};


int cLibrarian::addlexicon(sLanguage lang){
  PUSH("cLibrarian addlexicon");
  sLexLabel lxl;
  lxl.name = lang.name; lxl.kwrd = lang.kwrd;
  lxl.id = lib->stored();
  if  (lib->full()){ gmsg("Max. no. of lexicons exceeded."); POP; return -1; }
  lib->add(new cLexicon(lxl,this,gems));
  POP; return 0;
}

aGLDCell* cLibrarian::addradix(sExpression e){  
  PUSH("cLibrarian addradix");
  cLexicon* lex = chooselex(e.kwrd);
  
  if (lex){
    sReference r;
    r.lang = lex->id(); // hk, id of cell puts lex. unknown= -1
    r.id = NULL_ID; r.hk = NULL_HK;
    cRadix* rx = new cRadix(r,e.str,lex,gems);    
    lex->inscribe(rx);
    POP; return rx;
  } else {
    gmsg("Invalid language, lexicon not found!");
    POP; return NULL;
  }  
};

aGLDCell* cLibrarian::addcompound(sExpression e){
  PUSH("cLibrarian addcompound");
  cLexicon* lex = chooselex(e.kwrd);
  if (lex){
    sReference r;
    r.lang = lex->id(); // hk, id of cell puts lex. unknown= -1
    r.id = NULL_ID; r.hk = NULL_HK;
    // search for compound's EXISTING radixes:
    char* buf = strdup(e.str);
    char** bp = &buf;
    char* token;
    sExpression enode; strcpy(enode.kwrd,e.kwrd);
    cRefContainer* rc = new cRefContainer(GLDCELL_REL_DEFSIZE,GLDCELL_REL_PUFF,gems); // puff=5 cells
    while (token = strsep(bp, COMPOUND_DELIMITER)){
      strcpy(enode.str,token);
      aGLDCell* cnode = searchone(enode);
      
      if (!cnode) {
        gmsgs("Compound node (radix) not found:",token);
        gmsgs("radix language:",enode.kwrd);
        gmsg("Compound addition stopped!");
        /*delete buf;*/ delete rc; POP; return NULL;
      }
      if (cnode->gldctype()!=GLD_CT_RADIX){
        gmsgs("Compound node is not a basic type:radix but Clause or Compound:",token);
        gmsgs("compound/clause language:",enode.kwrd);
        gmsg("Compound addition stopped!");
        /*delete buf;*/ delete rc; POP; return NULL;
      }
      // all okay
      //gmsgs("TOKEN:",token);
      rc->add(cnode->ref());
    }
    cCompound* compx = new cCompound(r,rc,lex,gems);
    lex->inscribe(compx);
    //gmsg("FREEING:");    
    /*delete buf;*/ delete rc; 
    //gmsgs("COMPOUND:",compx->text(TF_PLAIN));
    POP; return compx;
  } else {
    gmsg("Invalid language, lexicon not found!");
    POP; return NULL;
  }
};

aGLDCell* cLibrarian::addclause(sExpression e){
  PUSH("cLibrarian addclause");
  cLexicon* lex = chooselex(e.kwrd);
  if (lex){
    sReference r;
    r.lang = lex->id(); // hk, id of cell puts lex. unknown= -1
    r.id = NULL_ID; r.hk = NULL_HK;
    // search for clause's EXISTING compounds and radixes:
    char* buf = strdup(e.str);
    char** bp = &buf;
    char* token;
    sExpression enode; strcpy(enode.kwrd,e.kwrd);
    cRefContainer* rc = new cRefContainer(GLDCELL_REL_DEFSIZE,GLDCELL_REL_PUFF,gems); // puff=5 cells
    while (token = strsep(bp, CLAUSE_DELIMITER)){
      strcpy(enode.str,token);
      aGLDCell* cnode = searchone(enode);
      
      if (!cnode) {
        gmsgs("Clause node (compound/radix) not found:",token);
        gmsgs("comp/radix language:",enode.kwrd);
        gmsg("Clause addition stopped!");
        /*delete buf;*/ delete rc; POP; return NULL;
      }
      if (cnode->gldctype()==GLD_CT_CLAUSE){
        gmsgs("Clause node is not a basic type:clause can't contain clause",token);
        gmsgs("compound/radix language:",enode.kwrd);
        gmsg("Clause addition stopped!");
        /*delete buf;*/ delete rc; POP; return NULL;
      }
      // all okay
      //gmsgs("TOKEN:",token);
      rc->add(cnode->ref());
    }
    cClause* clause = new cClause(r,rc,lex,gems);
    lex->inscribe(clause);
    //gmsg("FREEING:");    
    /*delete buf; ??????? delete rc;*/
    //gmsgs("CLAUSE:",clause->text(TF_PLAIN));
    POP; return clause;
  } else {
    gmsg("Invalid language, lexicon not found!");
    POP; return NULL;
  }
};


cCellBag* cLibrarian::search(sExpression e){
  PUSH("cLibrarian search");
  cCellBag* bag = new cCellBag(gems);
  cLexicon* lex = chooselex(e.kwrd);
  if (lex){
    //gmsgs("first search:",e.str);
    aGLDCell* cell = lex->texsea(e.str,TSM_FROM_START);
    //gmsg("clone");
    if (cell) { 
      cell = cell->clone(); bag->adds(cell);
    } else { delete bag; POP; return NULL; }
    //gmsg("browse search..");
    //gmsg(cell->text(TF_PLAIN));
    while (cell = lex->texsea(e.str,TSM_BROWSE)){
      protect(cell){ cell = cell->clone(); bag->adds(cell); }
    }  
    POP; return bag; //
  } else {
    gmsg("Invalid language, lexicon not found!");
    delete bag; POP; return NULL;
  }
};


inline int cLibrarian::addrelation(sRelation rel){
  PUSH("cLibrarian addrelation");
  if  (relist->full()){ gmsg("Max. no. of relations exceeded."); POP; return -1; }
  relist->add(rel);
  POP;return 0;
};

int cLibrarian::addrelation(sExpression e1, sExpRelation r, sExpression e2){
  PUSH("cLibrarian addrelation");
  static aGLDCell* c1=NULL;
  static sExpression lastexp = { '\0', '\0' };
  if (strcmp(e1.str,lastexp.str) || strcmp(e1.kwrd,lastexp.kwrd)) {
    c1 = searchone(e1);
    // cpy for next relation with same e1
    strcpy(lastexp.str,e1.str);
    strcpy(lastexp.kwrd,e1.kwrd);
  }
  aGLDCell* c2 = searchone(e2);
  // now put relation:
  if (!c2) { gmsgs("e2 not found:",e2.str); POP; return -1; }
  if (!c1) { gmsgs("e1 not found:",e1.str); POP; return -1; }
  RELATIONID rid;
  if ((rid = relkeytoid(r.kwrd)) < 0) {
    gmsgs("relation not found in rel.table:",r.kwrd);
    POP; return -1;
  }
  cRelation r1,r2;

  r1.rex.hk=NULL_HK;r1.rex.id=NULL_ID;r1.rex.lang=NULL_LANG;
  r2.rex.hk=NULL_HK;r2.rex.id=NULL_ID;r2.rex.lang=NULL_LANG;

  r1.rid = rid;  r2.rid = rid;
  // extensions of relations.
  if (!relisunary(rid) && strlen(r.e1EXTe2)){
    sExpression rexp; strcpy(rexp.str,r.e1EXTe2); strcpy(rexp.kwrd,e2.kwrd);
    //gmsg("rel clause1");
    aGLDCell* rex1 = searchone(rexp);
    if (!rex1) {
      gmsgs("incomplete extension of relation, relation add stopped:",rexp.str);
      gmsgs("language:",rexp.kwrd);
      POP; return -1;
    }
    //char* a=rex1->text(TF_PLAIN);
    //gmsgs("X",a); delete[] a;
    r1.rex = rex1->ref();
  }

  if (strlen(r.e2EXTe1)) {
    sExpression rexp; strcpy(rexp.str,r.e2EXTe1); strcpy(rexp.kwrd,e1.kwrd);
    //gmsg("rel clause2");
    aGLDCell* rex2 = searchone(rexp);
    if (!rex2) {
      gmsgs("incomplete extension of relation, relation add stopped:",rexp.str);
      gmsgs("language:",rexp.kwrd);
      POP; return -1;
    }
    //char* a=rex2->text(TF_PLAIN);
    //gmsgs("X",a); delete[] a;
    r2.rex = rex2->ref();
  }

  r1.ref = c1->ref(); r2.ref = c2->ref();
  //gmsgi("REL1:",r1.ref.lang);
  //gmsgi("REL2:",r1.ref.lang);
  //gmsgi("RID:",rid);
  success(c1->addrel(r2));
  if (!relisunary(rid)) success(c2->addrel(r1));
  POP; return 0;
};

void cLibrarian::printall(){
  PUSH("cLibrarian printall");
  // lexicons...$
  for (int i=0; i< lib->stored() ;i++){
    cLexicon*  lex = lib->read(i);
    printf("\n ~ LEXICON:%s \n", lex->name());
    lex->printall();
  }
  POP; return;
};

cRelationBag* cLibrarian::getrelation(aGLDCell* cell){
  PUSH("cLibrarian getrelation");
  protect(cell){    
    cRelRefBag* refbag = cell->getrelation();
    if (!refbag) { POP; return NULL; } // no relations found    
    cRelationBag* bag = new cRelationBag(refbag->stored(),gems);
    refbag->zeroindex();
    //if (refbag->stored()<1) gmsg("REFBAG empty!");    
    while (!refbag->over()) {
      // read
      cRelation r = refbag->browse();
      // data conversion
      //gmsg("GR4");
      sRelCard rc;      
       //gmsgi("REFLANG:",r.ref.lang);
       //gmsgi("REFID:",r.rid);$
       rc.inrelation = searchoneref(r.ref); // cell in relation with cell
       //gmsg("GR5");
       rc.condition = searchoneref(r.rex);  // relation conditions       
       protect(rc.inrelation) {         
           // name, keyword...
           sRelation rlist = relist->read(r.rid);
           strncpy(rc.name,rlist.name,LEX_NAME_LEN);
           strncpy(rc.kwrd,rlist.kwrd,LEX_KEYW_LEN);
           // add
           bag->add(rc);
           //gmsg("GR6"); 	      
       }     
    }    
    delete refbag;
    POP; return bag;
  }
  POP; return NULL;
}


/* text string search in chain */
aGLDCell* cKeyList::texsea(char* s, TSM scope){
  PUSH("cKeyList texsea");
  // stores # of recursive called texseas, if called texseas don't use
  // back query call and returns NOT found.
  static short int called = 0;
  // text search is allowed to call only once!
  if (called) { gmsg("texsea called twice, search forbidden!");
    POP; return NULL;
  }
  // increment # calls
  called++;
  if (chain->cells()<1) {
    //gmsg("No cells found in chain, internal error!");
    called--; POP; return NULL;
  }
  // search text string...
  // start is not affected texsea searches from last position!
  aGLDCell* cell;
  char* str;
  long int len;
  //gmsgi("SEARCH",scope);
  switch (scope) {
    case TSM_FS_EQU:{
      //gmsg("TSM_FS_EQU method");
      str = new char[MAX_STRING_LEN+1];// str[0] = '\0';
      len = MAX_STRING_LEN;
      chain->index(0); //back
      while (cell = chain->browse(1)){
        // GLD cell needs to call refsea() because of compound & clause!
        cell->text(str, TF_PLAIN, len);
        if (!strcmp(s,str)) { //gmsg("FOUND");
          chain->index(0); //back
          called--; delete[] str;
          POP; return cell;
        } // found
      }
      chain->index(0); //back
      delete[] str;
    } break;

    case TSM_START_KEEP:{
      int len = strlen(s); // search up to max. length of string
      str = new char[len+1]; //str[0] = '\0';
      long int store;
      store = chain->index();
      chain->index(0); // search from start
      while (cell = chain->browse(1)){
        // GLD cell needs to call refsea() because of compound & clause!
        cell->text(str, TF_PLAIN, len);
        if (!strcmp(str,s)) {
          // restore current index:
          chain->index(store);
          called--; delete[] str;
          POP; return cell;
        } // found
      }
      chain->index(store);
      delete[] str;
    } break;

    case TSM_FROM_START:{
      int len = strlen(s); // search up to max. length of string
      str = new char[len+1]; //str[0] = '\0';
      chain->index(0);
      while (cell = chain->browse(1)){
        // GLD cell needs to call refsea() because of compound & clause!
        //~ gmsgi("IndexBEFORE:",chain->index());
        cell->text(str, TF_PLAIN, len);
        // if (strlen(str)!=len) gmsgs("ERROR:",str);
        //~ gmsgi("IndexAFTER:",chain->index());
        if (!strcmp(str,s)) { //gmsg("FOUND");
          //chain->index(0); // back
          called--; delete[] str;
          POP; return cell;
        } // found
      }
      //chain->index(0); // back
      delete[] str;
    } break;
    case TSM_BROWSE:{
      int len = strlen(s); // search up to max. length of string
      str = new char[len+1]; str[0] = '\0';
      while (cell = chain->browse(1)){
        // GLD cell needs to call refsea() because of compound & clause!
        //~ gmsgi("IndexBEFORE:",chain->index());
        cell->text(str, TF_PLAIN, len);
	// if (strlen(str)!=len) gmsgs("ERROR:",str);
        if (!strcmp(str,s)) { //gmsg("FOUND");
          called--; delete[] str;
          POP; return cell;
        } // found
      }
      //chain->index(0); // back
      delete[] str;
    } break;
  }
  
  called--;
  POP;
  return NULL;
};

/* reference search in chain */
aGLDCell* cKeyList::refsea(sReference ref){
  PUSH("cKeyList refsea");
  //gmsg("entering REFsearch");
  // stores # of recursive called texseas, if called texseas don't use
  // back query call and return NOT found.
  static short int called = 0;
  // text search is allowed to call only once!
  if (called>1) { gmsg("refsea called 3x, search forbidden!");
    POP; return NULL;
  }
  // increment # calls
  called++;
  // search reference...
  aGLDCell* cell;
  // check if key ~~ backqry->reftokey(ref)  
  /* NOT USED NOW - obsolete
  if (backquery){
    if (key!=backquery->reftokey(ref))
    
      // not found in this chain, try back query hashSearch
      gmsg("Back query call by refsea KeyList");
      cell = backquery->refsea(ref);
      called--; POP; return cell;
  } else {
      gmsg("Back query NULL!");
      called--; POP; return NULL;
    }*/
    
  // key is okay, search in this chain...
  // start is not affected refsea searches from last position!
  // store current index:
  long int store=chain->index();
  success(chain->index(0));
  while (cell = chain->browse(1)){
    // GLD cell needs to call refsea() because of compound & clause!
    //gmsgi("!!! INDEX !!! ",chain->index());
    sReference r = cell->ref();
    //gmsgi("TAG!",chain->cells());
    //gmsgs("TAG INDEX!",);
    if ((r.lang==ref.lang) && (r.id==ref.id) && (r.hk==ref.hk)){
      // restore current index:
      //gmsgi(">>>>>>FOUND ID:",r.id);
      chain->index(store);
      called--; POP; return cell;
    } // found
  }
  //success(chain->index(0));
  // not found
  // restore current index:
  //gmsgi("RESTORE",store);

  chain->index(store);
  called--; POP; return NULL;
};

void cClause::text(char* str, TEXTFORMAT tf, int chars){
  PUSH("cClause text1");
  if (!backqry) gmsg("Compound require back query call!"); else {
    rprotect(1,chars,MAX_STRING_LEN){
      protect(str){
        str[0]='\0'; //blank sring
        // assemble compoud from radixes:
        clause->zeroindex();
        // if text format bla|bla|ble required
        // allocate +(comps->stored()-1) chars.
        int rest = chars; // not required +comps->stored..
        char* fragment = new char[chars+1];
	//gmsgi("#Chars:",chars);
        while ((!clause->over()) && (rest>0)){
	  //gmsgi("Rest:",rest);
          // get one radix:
          sReference cref = clause->browse();
          // search for radix:
          protect(backqry){
            aGLDCell* cell = backqry->refsea(cref);
             //~ gmsgi("cref.id",cref.id);
             //~ gmsgi("cref.hk",cref.hk);
             //~ gmsgi("cref.lang",cref.lang);
             // get content of radix + add:
            protect(cell){
              // check if cell is radix!
              if (cell->gldctype()== GLD_CT_CLAUSE)
                gmsg("Radix/Compound expected, clause can't contain clause "
                      ", error in database!");
              else {
                cell->text(fragment,TF_PLAIN,rest);
                //gmsg(fragment);
                //char* s = str;
                //char* r = radix;
                //while ((*s++ = *r++)&&((radix-r)<=rest));
                strncat(str,fragment,rest);		
                rest-= strlen(fragment);
		if (!clause->over() && (rest>0)){
		  strcat(str,CLAUSE_SEPARATOR);
		  rest-=1;
		}  
                // if text format bla|bla|ble required
                // add '|' strcpy(str,"|");
                // dont' rest++ !
              }
            }
          }
        }
	//gmsgs("X:",str);
        delete[] fragment;
      }
    }
  }
  POP; return;  
};

char* cClause::text(TEXTFORMAT tf){
  PUSH("cClause text2");
  //~gmsg("cCompound asked for text1");
  if (!backqry) gmsg("Clause require back query call!"); else {
    // allocate string container.
    int ncells = clause->stored();
    cStrContainer* scont = new cStrContainer(ncells,gems);
    // assemble compoud from radixes:
    clause->zeroindex();
    // if text format bla|bla|ble required
    int len = 0; // total lenght of contacted string
    while (!clause->over()){
      // get one radix:
      sReference cref = clause->browse();
      //gmsgi("REF HK:",cref.hk);
      //gmsgi("REF LANG:",cref.lang);
      //gmsgi("REF ID:",cref.id);
      // search for radix:
      aGLDCell* cell = backqry->refsea(cref);
      protect(cell){
        // check if cell is radix!
        if (cell->gldctype()== GLD_CT_CLAUSE)
          gmsg("Radix/compound expected, clause can't contain clause "
                ", error in database!");
        else {
          char* s = cell->text(TF_PLAIN);
          //~gmsg(s);
          len += strlen(s); len+=1;
          scont->add(s);
        }
      }
    }
    char* outs = new char[len+1];
    // contact partial strings (radixes):
    scont->zeroindex();
    outs[0] = '\0'; // blank string
    for (int i = 0; i<ncells; i++) {
      char* tmp =  scont->browse();
      strcat(outs,tmp);
      if (i<(ncells-1)) strcat(outs,CLAUSE_SEPARATOR);
      delete[] tmp;
    };    
    delete scont;
    POP; return outs;
  }
  POP;
  // char* str = new char[strlen(radix)+1];
  // strcpy(str,radix);
  return NULL;
};
